// Copyright 2003 Adam Megacz, see the COPYING file for licensing [LGPL]
// Author: Brian Alliet

package org.xwt.plat;
import org.xwt.*;
import org.xwt.util.*;

abstract class OpenGL {
    static final boolean pretendToBeACrappyVideoCard = false;
    boolean rectangularTextures;
    boolean hasRectangularTextures() { return rectangularTextures; }
    int maxTexSize;
    int maxRectTexSize;
    String version;
    String renderer;
    String vendor;
    
    private native void natInit();
    
    private static float parseVersion(String s) {
        int end = s.indexOf(' ');
        if(end < 0) end = s.length();
        try {
            return Float.parseFloat(s.substring(0,end));
        } catch(NumberFormatException e) {
            return 1.1f; // just a guess
        }
    }
    // This MUST be called after OpenGL is instansiated (and activateSharedContext is functioning)
    public void init() throws NotSupportedException {
        natInit();
        float v = parseVersion(version);
        // If we disable linear filtering (and therefor GL_CLAMP_TO_EDGE) we could probably get by with less
        if(v < 1.2) throw new NotSupportedException("OpenGL 1.2 or greater is required. (you have: " + version +" - " + v + ")");
        if(pretendToBeACrappyVideoCard) {
            maxTexSize = 512;
            maxRectTexSize = 0;
            rectangularTextures = false;
        }
        Log.log(this,"Renderer: " + renderer);
        Log.log(this,"Version: " + version);
        Log.log(this,"Vendor: " + vendor);
        Log.log(this,"Rectangular textures: " + (rectangularTextures ? "supported" : "unsupported"));
        Log.log(this,"Max texture size: " + maxTexSize);
        Log.log(this,"Max rectangular texture size: " + maxRectTexSize);
    }
    
    protected abstract void activateSharedContext();

    public static class NotSupportedException extends Exception {
        public NotSupportedException(String s) { super(s); }
    }

    public static abstract class GLDoubleBuffer extends DoubleBuffer {
        protected int width;
        protected int height;
        public int getWidth() { return width; }
        public int getHeight() { return height; }
        
        private boolean glScissorEnabled = false;
        
        public GLDoubleBuffer(int width, int height) {
            this.width = width;
            this.height = height;
        }
        
        // This should activate the drawing context and prepare the double buffer for drawing
        protected abstract void activateContext();
        
        protected static native void drawableInit(int w, int h);
        
        private static native void setColor(int color);
        
        public native void setClip(int x, int y, int x2, int y2);
        public native void resetClip();
        public native void fillRect(int x, int y, int x2, int y2, int color);
        public native void fillTrapezoid(int x1, int x2, int y1, int x3, int x4, int y2, int color);
            
        public void drawString(String font, String text, int x, int y, int color) {
            //System.out.println("drawString(): " + text);
        }
        
        public void drawPicture(org.xwt.Picture source, int x, int y) {
            activateContext();
            GLPicture p = (GLPicture) source;
            p.draw(x,y);
        }
        public void drawPicture(org.xwt.Picture source, int dx1, int dy1, int dx2, int dy2,int sx1, int sy1, int sx2, int sy2) {
            activateContext();
            GLPicture p = (GLPicture) source;
            p.draw(dx1,dy1,dx2,dy2,sx1,sy1,sx2,sy2);
        }
    }
    
    public final static int roundToPowerOf2(int n) {
        if(((n-1)&n)==0) return n;
        for(int x=2;x!=0;x<<=1) if(n < x) return x;
        return 0;
    }
        
    public Picture createPicture(int[] data, int w, int h) {
        if(w*h != data.length) throw new Error("w*h != data.length");
        if(rectangularTextures && w <= maxRectTexSize && h <= maxRectTexSize) new RectGLPicture(data,w,h,this);
        if(w <= maxTexSize && h <= maxTexSize) return new SquareGLPicture(data,w,h,this);
        return new MosaicGLPicture(data,w,h,this);
    }
    
    private native void natDeleteTexture(int tex);
    public void deleteTexture(final int tex) {
        // CHECKME: Is this safe to do from finalize()?
        // natDeleteTexture MUST be run from the message queue thread
        MessageQueue.add(new Message() { public void perform() {
            natDeleteTexture(tex);
        }});
    }
    
    private static abstract class GLPicture extends Picture {
        protected int width;
        protected int height;
        
        public final int getWidth() { return width; }
        public final int getHeight() { return height; }
                
        public GLPicture(int w, int h) {
            this.width = w;
            this.height = h;
        }
        
        public void draw(int x, int y) { draw(x,y,x+width,y+height,0,0,width,height); }
        public abstract void draw(int dx1, int dy1, int dx2, int dy2,int sx1, int sy1, int sx2, int sy2);
        protected abstract void finalize();
    }
    
    private static class RectGLPicture extends GLPicture {
        private OpenGL gl;
        public int textureName;
                
        public native void natInit(int[] data);
        
        public RectGLPicture(int[] data,int w, int h, OpenGL gl) {
            super(w,h);
            this.gl = gl;
            natInit(data);
        }
    
        public native void draw(int dx1, int dy1, int dx2, int dy2,int sx1, int sy1, int sx2, int sy2);  
        protected void finalize() { gl.deleteTexture(textureName); }
    }
    
    private static class SquareGLPicture extends GLPicture {
        private int texHeight;
        private int texWidth;
        private OpenGL gl;
        public int textureName;
                
        public native void natInit(int[] data);
                
        public SquareGLPicture(int[] data,int w, int h, OpenGL gl) {
            super(w,h);
            this.gl = gl;
            if(w > 0x7fffffff || h > 0x7fffffff) throw new Error("insane texture size: " + w + "x" + h);
            texHeight = roundToPowerOf2(h);
            texWidth = roundToPowerOf2(w);
            natInit(data);
        }
        
        public native void draw(int dx1, int dy1, int dx2, int dy2,int sx1, int sy1, int sx2, int sy2); 
        protected void finalize() { gl.deleteTexture(textureName); }
    }
    
    private static class MosaicGLPicture extends GLPicture {
        private static final int overlap = 8;
        
        int psize;
        GLPicture[][] pics;
        
        private static float wastedSpace(int w,int h,int size) {
            int w2 = size * ((w+size-1)/size);
            int h2 = size * ((h+size-1)/size);
            return 1.0f-(float)(w*h)/(float)(w2*h2);
        }
        
        private static int[] subData(int[] data, int x, int y, int w, int h, int rowSize) {
            int[] sub = new int[w*h];
            int row = y;
            try {
            for(int i=0;i<h;i++,row++)
                for(int j=0;j<w;j++)
                    sub[i*w+j] = data[row*rowSize+j+x];
            } catch(ArrayIndexOutOfBoundsException e) {
                Log.log(MosaicGLPicture.class,"subData(data("+data.length+"),"+x+","+y+","+w+","+h+","+rowSize);
                throw e;
            }
            return sub;
        }
        
        public MosaicGLPicture(int[] data,int w, int h, OpenGL gl) {
            super(w,h);
            psize = gl.maxTexSize;
            while(wastedSpace(w,h,psize) > 0.40) psize/=2;
            Log.log(this,"Using psize: " + psize + " for " + w + "x" + h + " image (wasted space: " + wastedSpace(w,h,psize));
            
            int rows = (h+psize-1)/psize;
            int cols = (w+psize-1)/psize;
            int tmp,tmp2;
            
            pics = new GLPicture[rows][cols];
            for(int i=0;i<rows-1;i++)
                for(int j=0;j<cols-1;j++)
                    pics[i][j] = new SquareGLPicture(subData(data,j*psize,i*psize,psize,psize,w),psize,psize,gl);
            tmp = (rows-1)*psize;
            for(int i=0;i<cols-1;i++)
                pics[rows-1][i] = new SquareGLPicture(subData(data,i*psize,tmp,psize,h-tmp,w),psize,h-tmp,gl);
            tmp2 = (cols-1)*psize;
            for(int i=0;i<rows-1;i++)
                pics[i][cols-1] = new SquareGLPicture(subData(data,tmp2,i*psize,w-tmp2,psize,w),w-tmp2,psize,gl);
            pics[rows-1][cols-1] = new SquareGLPicture(subData(data,tmp2,tmp,w-tmp2,h-tmp,w),w-tmp2,h-tmp,gl);
        }
        protected void finalize() {  }
        
        private static final int max(int a, int b) { return a > b ? a : b; }
        private static final int min(int a, int b) { return a < b ? a : b; }
        
        public void draw(int dx1, int dy1, int dx2, int dy2,int sx1, int sy1, int sx2, int sy2) {
            double xscale = (double)(dx2-dx1)/(double)(sx2-sx1);
            double yscale = (double)(dy2-dy1)/(double)(sy2-sy1);
            int totalWidth = width;
            int totalHeight = height;
            // *{x,y}{1,2} key: d=dest s=src, p=bounds of this picture, i=intersection of s and p, pd = dest of this pic
            //System.out.println("Starting draw..." + sx1 + "," + sy1 + " " + sx2 + "," + sy2 + " to " + dx1 +"," + dy1 + " " + dx2 + "," + dy2);
            for(int i=0;i<pics.length;i++) {
                for(int j=0;j<pics[i].length;j++) {
                    int px1 = j*psize;
                    int py1 = i*psize;
                    int px2 = min(px1+psize,totalWidth);
                    int py2 = min(py1+psize,totalHeight);
                    int ix1 = max(px1,sx1);
                    int iy1 = max(py1,sy1);
                    int ix2 = min(px2,sx2);
                    int iy2 = min(py2,sy2);
                    if(ix1 >= ix2 || iy1 >= iy2) continue; // no intersection
                    
                    int pdx1 = dx1 + (int) (xscale*(ix1-sx1));
                    int pdy1 = dy1 + (int) (yscale*(iy1-sy1));
                    int pdx2 = dx2 - (int) (xscale*(sx2-ix2));
                    int pdy2 = dy2 - (int) (yscale*(sy2-iy2));
                    
                    //System.out.println("" + i + "," + j + " is good... drawing from " + (ix1-px1) + "," + (iy1-py1) + " " + (ix2-px1) + "," + (iy2-py1) + " to " + pdx1 + "," + pdy1 + " to " + pdx2 + "," + pdy2);
                    
                    pics[i][j].draw(pdx1,pdy1,pdx2,pdy2,ix1-px1,iy1-py1,ix2-px1,iy2-py1);
                }
            }
        }
    }
}
